# Errors and Exceptions

class TranslatorException(Exception):
    """Base exception to be able to catch all exceptions in this module"""
    pass


class ProviderAlreadyAdded(TranslatorException):
    def __init__(self, name):
        super().__init__(f"A provider with the name '{name}' has already been added.")


class LanguageNotSupported(TranslatorException):
    def __init__(self, language):
        super().__init__(f"Language {language} is not supported with the current providers.")



import types
from typing import Optional, Dict

from .abc import BaseProvider, Translation
from .errors import ProviderAlreadyAdded, LanguageNotSupported
from .caseinsensitivedict import CaseInsensitiveDict


class AsyncTranslate:
    def __init__(self):
        self._languages = CaseInsensitiveDict()
        self._providers = CaseInsensitiveDict()
        self.language_names: Dict[str, str] = {}

    @property
    def providers(self):
        """Returns read-only copy of the providers"""
        return types.MappingProxyType(self._providers)

    @property
    def default_provider(self) -> BaseProvider:
        """Returns the default (first) provider"""
        return next(iter(self._providers.values()))

    @property
    def languages(self):
        """Returns read-only copy of the supported languages"""
        return types.MappingProxyType(self._languages)

    async def add_provider(self, provider: BaseProvider):
        """Add a translator provider"""
        if not isinstance(provider, BaseProvider):
            raise TypeError('Providers must derive from BaseProvider')

        name = provider.name

        if name in self._providers:
            raise ProviderAlreadyAdded(name)
        self._providers[name] = provider

        for lang, lang_name in (await provider.get_languages()).items():
            bck_str = [name]
            if added_lang := self._languages.get(lang):
                self._languages[lang] = added_lang + bck_str
            else:
                self._languages[lang] = bck_str
                # Only set name if it hasn't been added already
                self.language_names[lang] = lang_name

    async def add_providers(self, *backends: [BaseProvider]):
        """Add multiple providers"""
        for b in backends:
            await self.add_provider(b)

    def provider_for(self, language: str, preferred: Optional[str] = "") -> BaseProvider:
        """Returns a valid provider for the specified language"""
        # precondition self._providers has at least one provider in it
        preferred = preferred.casefold()
        try:
            providers = map(lambda x: x.casefold(), self._languages.get(language))
        except TypeError:
            raise LanguageNotSupported(language)
        return self._providers[preferred] if preferred in providers else self._providers[providers[0]]

    async def detect(self, content: str, preferred: Optional[str] = None) -> (str, BaseProvider):
        provider: BaseProvider = self.providers[preferred] if preferred else self.default_provider
        return await provider.detect(content), provider

    async def translate(self, content: str, to: str, fro="", preferred: Optional[str] = None, **options) -> [Translation]:
        """
        Translates the given content to the specified language
        :param content: Content to translate
        :param to: Language code to translate to
        :param fro: Optional language to translate from
        :param preferred: Optional provider to prefer
        :return: A list of translations
        """
        provider = self.provider_for(to, preferred=preferred)
        return await provider.translate(content, to=to, fro=fro, **options)



import os
import uuid
from typing import Dict

import aiohttp
from aiocache import cached

from async_translate.abc import BaseProvider, ONE_DAY, Translation

MS_API_VER = "?api-version=3.0"


class Azure(BaseProvider):
    """
    Microsoft Azure Cognitive Services translator Backend
    Required environment variables:
    `MS_TRANSLATE_KEY`
    """
    backend = "azure"
    ms_endpoint = "https://api.cognitive.microsofttranslator.com/"
    icon = "https://connectoricons-prod.azureedge.net/microsofttranslator/icon_1.0.1303.1871.png"

    def __init__(self):
        self.session = aiohttp.ClientSession()
        if 'MS_TRANSLATE_KEY' not in os.environ:
            raise Exception("Please set/export the 'MS_TRANSLATE_KEY' environment variable.")
        self.ms_key = os.environ['MS_TRANSLATE_KEY']

    @property
    def headers(self):
        return {
            'Ocp-Apim-Subscription-Key': self.ms_key,
            'Content-type': 'application/json',
            'X-ClientTraceId': str(uuid.uuid4())
        }

    @cached(ttl=ONE_DAY)
    async def get_languages(self) -> Dict[str, str]:
        url = self.ms_endpoint + "languages" + MS_API_VER + "&scope=translation"
        async with self.session.get(url, headers=self.headers) as resp:
            raw_langs = (await resp.json())['translation']

            return {key: value['name'] for key, value in raw_langs.items()}

    async def detect(self, content) -> str:
        """Detect the language of the given content"""
        url = self.ms_endpoint + "detect" + MS_API_VER
        json_content = [{"text": content}]
        async with self.session.post(url, headers=self.headers, json=json_content) as resp:

            return (await resp.json())[0]['language']

    async def translate(self, content: str, to: str, fro="", **options) -> [Translation]:
        """
        Translates text using the Azure API
        :param content: the text to translate
        :param to: the language code to translate to
        :param fro: Optional language to translate from
        :return: the translation
        """
        url = self.ms_endpoint + "translate" + MS_API_VER + "&to=" + to
        if fro:
            url += "&from=" + fro

        profanity_filter = options.get('profanity_filter') or 0

        url += "&profanityAction=" + ['NoAction', 'Marked', 'Marked'][profanity_filter]
        if profanity_filter == 1:
            url += "&profanityMarker=Tag"

        json_content = [{"text": content}]
        async with self.session.post(url, headers=self.headers, json=json_content) as resp:
            json_response = (await resp.json())[0]
            translation_response = []

            try:
                detected_language = json_response.get('detectedLanguage').get('language')
            except AttributeError:
                detected_language = None

            if ts := json_response.get('translations'):
                for translation in ts:
                    translation_response.append(
                        Translation(text=translation.get('text'), to=translation.get('to'),
                                    provider=self,
                                    detectedLanguage=detected_language)
                    )
            return translation_response



from html import unescape
from typing import Dict

from aiocache import cached
from google.cloud.translate_v3.services.translation_service.async_client import TranslationServiceAsyncClient

from async_translate.abc import BaseProvider, ONE_DAY, Translation


class Google(BaseProvider):
    icon = "https://i.imgur.com/jDPXiQh.png"

    def __init__(self):
        self.client = TranslationServiceAsyncClient()
        self.parent = "projects/mr-translate-1577912381600/locations/global"

    @cached(ttl=ONE_DAY)
    async def get_languages(self, display_language_code="en") -> Dict[str, str]:
        langs = (await self.client.get_supported_languages(parent=self.parent,
                                                           display_language_code=display_language_code)).languages
        return {lang.language_code: lang.display_name
                for lang in
                filter(lambda l: l.support_source and l.support_target, langs)
                }

    async def detect(self, content) -> str:
        res = (await self.client.detect_language(parent=self.parent, content=content)).languages[0]
        return res.language_code

    async def translate(self, content: str, to: str, fro="", **options) -> [Translation]:
        """
        Translates text using the Google API
        :param content: the text to translate
        :param to: the language code to translate to
        :param fro: Optional language to translate from
        :return: the translation
        """
        params = {"contents": [content],
                  "target_language_code": to,
                  "parent": self.parent}
        if fro:
            params['source_language_code'] = fro

        translation_response = []
        translations = (await self.client.translate_text(**params)).translations
        for trans in translations:
            translation_response.append(Translation(
                to=to, text=unescape(trans.translated_text),
                provider=self,
                detectedLanguage=trans.detected_language_code or None
            ))
        return translation_response
